require 'morq/base'
require 'morq/list'
require 'morq/query'
require 'morq/spam'

# READ(mid):
# - MARK_READ(mid)
# - return path (mid[3]/mid[0]/mid)

# MARK_READ_ON/MARK_FLAG_ON(mid):
# - get index(mid)
# - index->flags |= FLAG (READ or FLAG)
# - upadte index

# MARK_READ_OFF/MARK_FLAG_OFF(mid):
# - get index(mid)
# - index->flags &= ~FLAG (READ or FLAG)
# - upadte index

module Morq::Action
  include Morq::Common
  include Morq::List
  include Morq::Query
  include Morq::Spam

  READ_MARK = 1
  PENDING_MARK = 2
  REPLIED_MARK = 4

  def turn_on_flag(mid, flag)
    pmid = pack_id(mid)
    open_morq_db("index", :update) do |db|
      index = unpack_index(db[pmid])
      index[0] |= flag
      db[pmid] = pack_index(index)
    end
  end

  def turn_off_flag(mid, flag)
    pmid = pack_id(mid)
    open_morq_db("index", :update) do |db|
      index = unpack_index(db[pmid])
      index[0] &= ~flag
      db[pmid] = pack_index(index)
    end
  end

  def mark_as_read(mid)
    turn_on_flag(mid, READ_MARK)
  end
  def mark_as_unread(mid)
    turn_off_flag(mid, READ_MARK)
  end

  def mark_as_replied(mid)
    turn_on_flag(mid, REPLIED_MARK)
  end
  def mark_as_unreplied(mid)
    turn_off_flag(mid, REPLIED_MARK)
  end

  def mark_pending(mid)
    add_label(mid, "pending")
    turn_on_flag(mid, PENDING_MARK)
  end
  def mark_unpending(mid)
    remove_label(mid, "pending")
    turn_off_flag(mid, PENDING_MARK)
  end

  def delete_forever(mid)
    dbpath = morq_file("index")
    delete_inverted_index(mid)
    pmid = pack_id(mid)
    open_morq_db("index", :update) do |db|
      db.delete(pmid)
    end
    File.unlink(path)
    open_thread_db2() do |db1, db2|
      db1.delete(pmid)
    end
  end

  def purge_db()
    system("estindex purge #{morq_file("index")} > /dev/null")
  end

  def optimize_db()
    system("estindex purge #{morq_file("index")} > /dev/null")
    system("estindex optimize #{morq_file("index")} > /dev/null")
    system("estindex relate #{morq_file("index")} > /dev/null")
  end

  def show_inbox(n=0, unit=0)
    show_label("inbox")
  end
  def show_label(label, n=0, unit=0)
    box_label(label).list
  end
  def show_thread(tid)
    box_thread(tid).list
  end
  def show_query(query)
    box_query(query).list
  end
  def show_similarity(mid)
    box_similarity(mid).list
  end

  def show_keywords(mid)
    raise RuntimeError, "Rast index does not support keyword list"
    dbpath = morq_file("index");
    puts Morq::Estraier::estkeywords(dbpath, File.join(dbpath, "_score"), eid_from_mid(mid)).join(" ")
  end

  def show_mail(mid)
    mark_as_read(mid)
    path = mid_path(mid)
    open(path) do |f|
      f.each do |line|
        print line
      end
    end
  end

  def archive_mail(mid)
    open_label_db(:update){|db| db.remove(mid, "inbox")}
  end
  def archive_read(label="inbox")
    open_label_db(:update) do |db|
      data = db.mid_data(label)
      open_morq_db("index") do |idx|
        list_each(data) do |elem|
          begin
            index = unpack_index(idx[elem])
          rescue
            db.remove(unpack_id(elem), "inbox")
            next
          end
          next if (index[0] & (READ_MARK|PENDING_MARK)) != READ_MARK
          db.remove(unpack_id(elem), "inbox")
        end
      end
    end
  end

  def archive_thread(tid)
    open_label_db() do |db|
      list_each(get_thread_list(tid)) do |elem|
        db.remove(unpack_id(elem), "inbox")
      end
    end
  end

  def add_label(mid, label)
    open_label_db(true) do |db|
      db.add(mid, label)
    end
  end
  def remove_label(mid, label)
    open_label_db(true) do |db|
      db.remove(mid, label)
    end
  end
  def kill_label(label)
    open_label_db(true) do |db|
      db.kill(label)
    end
  end
  def list_label(mid)
    labels = []
    open_label_db() do |db|
      if mid == ""            # list all labels
        db.list.each do |label|
          labels.push label
        end
      else
        for label in db.get_labels(mid.to_i)
          labels.push label
        end
      end
    end
    puts labels.join(", ")
  end
  def remove_labels(mid)
    open_label_db(true) do |db|
      labels = db.get_labels(mid)
      labels.each do |label|
        db.remove(mid, label)
      end
    end
  end

  def re_index(mid)
    require 'morq/inc'
    Morq::Inc.new.inverted_index_mid(mid)
  end
  def trash(mid)
    remove_labels(mid)
    add_label(mid, "trash")
  end
  def no_trash(mid)
    remove_label(mid, "trash")
    add_label(mid, "inbox")
#    re_index(mid)
  end
  def spam(mid)
    report_spam(mid_path(mid))
    remove_labels(mid)
    add_label(mid, "spam")
  end
  def no_spam(mid)
    report_nospam(mid_path(mid))
    remove_label(mid, "spam")
    add_label(mid, "inbox")
    re_index(mid)
  end

  def move_to_inbox(mid)
    if open_label_db() {|db|db.get_labels(mid)}.include?("spam")
      no_spam(mid)
    else
      no_trash(mid)
    end
  end

  def delete_all(label)
    open_label_db(:update) do |db|
      list_each(db.mid_data(label)) do |pmid|
        mid = unpack_id(pmid)
        delete_forever(mid) rescue nil
      end
      db.clear(label)
    end
  end
  def clear_trash()
    delete_all("trash")
  end
  def clear_spam()
    delete_all("spam")
  end

  def reindex(mid)
    require 'morq/inc'

    pmid = pack_id(mid)
    flag = 0
    begin
      open_morq_db("index") do |db|
        index = unpack_index(db[pmid])
        flag = index[0]
      end
    rescue
    end
    Morq::Inc.new.instance_eval{register_index(mid)}
    open_morq_db("index", :update) do |db|
      index = unpack_index(db[pmid])
      index[0] = flag
      db[pmid] = pack_index(index)
    end
  end

  def exec_input(cmd, rest)
    if rest
      rest.downcase!
      rest.strip!
    end
    case cmd
    when "show_inbox", "inbox"
      show_inbox()
    when "show_label"
      show_label(rest)
    when "show_thread"
      show_thread(rest.to_i)
    when "show_query", "query"
      show_query(rest)
    when "show_similarity", "similarity"
      show_similarity(rest.to_i)
    when "show_keywords", "keywords"
      show_keywords(rest.to_i)
    when "show_mail", "mail", "read"
      show_mail(rest.to_i)
    when /^\d+$/
      show_mail(cmd.to_i)
    when "mark_as_read"
      rest.split.each{|x| mark_as_read(x.to_i)}
    when "mark_as_unread"
      rest.split.each{|x| mark_as_unread(x.to_i)}
    when "mark_as_replied"
      mark_as_replied(rest.to_i)
    when "mark_as_unreplied"
      mark_as_unreplied(rest.to_i)
    when "mark_pending"
      mark_pending(rest.to_i)
    when "mark_unpending"
      mark_unpending(rest.to_i)
    when "delete_forever"
      delete_forever(rest.to_i)
    when "purge_db"
      purge_db()
    when "optimize_db"
      optimize_db()
    when "archive_mail", "archive"
      rest.split.each{|x|archive_mail(x.to_i)}
    when "archive_read"
      archive_read()
    when "arcihive_thread"
      archive_thread(rest.to_i)
    when "add_label"
      mid, label = rest.split
      add_label(mid.to_i, label)
    when "add_labels"
      label, *mids = rest.split
      mids.each{|x|add_label(x.to_i, label)}
    when "remove_label"
      mid, label = rest.split
      remove_label(mid.to_i, label)
    when "remove_labels"
      label, *mids = rest.split
      mids.each{|x|remove_label(x.to_i, label)}
    when "kill_label"
      kill_label(rest)
    when "list_label"
      list_label(rest)
    when "trash"
      rest.split.each{|x| trash(x.to_i)}
    when "no_trash"
      rest.split.each{|x| no_trash(x.to_i)}
    when "spam"
      rest.split.each{|x| spam(x.to_i)}
    when "no_spam"
      rest.split.each{|x| no_spam(x.to_i)}
    when "clear_trash"
      clear_trash()
    when "clear_spam"
      clear_spam()
    when "move_to_inbox"
      rest.split.each{|x| move_to_inbox(x.to_i)}
    when "reindex"
      reindex(rest.to_i)
    when "max_mid"
      p mid_max()
    when "path"
      puts mid_path(rest.to_i)
    else
      printf "NO COMMAND - %s\r\n", cmd
    end
#  rescue
    # ignore errors
  end
end

if __FILE__ == $0
  class Morq::App
    include Morq::Action

    def parse_input(line)
      cmd, rest = line.split(/ +/, 2)
      exec_input(cmd, rest)
    end

    def run()
      STDIN.each_line do |line|
        line.chomp!
        begin
          parse_input(line)
        rescue => e
          puts e.message
          puts e.backtrace
        ensure
          print ".\r\n"
        end
      end
    end
  end
  Morq::App.new.run
end
