require 'yaml'
require 'depot'		# QDBM
require 'fileutils'

module Morq
  Config = {
    'base-dir' => File.join(ENV["HOME"], ".morq")
  }

  module Common
    def morq_file(fname)
      File::join(Morq::Config["base-dir"], fname)
    end

    def mid_path(mid)
      File::join(Morq::Config["base-dir"], "spool", "%02x"%((mid>>12)&0x1f), "%02x"%(mid&0xff), "%d"%mid)
    end

    def mid_max
      path = morq_file("mid")
      open(path, "r+") do |f|
        f.flock(File::LOCK_EX)
        data = f.read
        if data.nil? or data == ""
          num = 0
        else
          num = unpack_id(data)
        end
        return num
      end
    end

    def open_morq_db(dbname, update=false, &block)
      path = morq_file("db")
      if update
        flags = Depot::OWRITER | Depot::OCREAT
        FileUtils.mkpath(path)
      else
        flags = Depot::OREADER
      end

      db = Depot::new(File::join(path, dbname), flags)
      if block
        begin
          yield db
        ensure
          db.close
        end
      else
        db
      end
    end

    def pack_id(id)
      [id].pack("I")
    end
    def unpack_id(pid)
      pid.unpack("I")[0]
    end
    def pack_index(idx)
      idx.pack("IIIIZ*Z*")
    end
    def unpack_index(idx)
      idx.unpack("IIIIZ*Z*")
    end

    def list_each(list)
      i = 0
      len = list.size
      while i < len
        elem = list[i,4]
        yield elem
        i += 4
      end
    end

    def list_include?(list, elem)
      list_each(list) do |e|
        if e == elem
          return true
        end
      end
      false
    end

    def list_delete(list, elem)
      i = 0
      len = list.size
      while i < len
        e = list[i,4]
        if e == elem
          list[i,4]=""
          return
        end
        i += 4
      end
    end
  end

  extend Morq::Common

  base = Config['base-dir']
  if File.exist?(morq_file("config"))
    open(morq_file("config")) do |f|
      config = YAML.load(f)
      Config.update(config)
    end
  end
end

# mail operation
#  common - read, flag, read, unread
#  INBOX - archive, spam, trash
#  SEARCH - inbox, spam, trash
#  TRASH - inbox, spam, remove forever
#  SPAM - inbox, no spam, remove forever

# INC: done - inc.rb
# - safecat(1) incoming mail
# - allocate morq-id (32bit int)
# - link incoming mail to path (mid[3]/mid[0]/mid)
# - update index database (mid -> flags, eid, tid, date, from, subject)
# - update thread database (message-id -> mid / mid -> tid / tid -> mid...)
# - run `echo path | estindex register -msuf "" -'
# - LABEL_ADD(mid, *ARGV)
# - LABEL_ADD(mid, inbox) unless ARGV include? "trash" nor "spam"
# - remove incoming file (link in path remains)

# SHOW_INBOX: done
# - LIST_LABEL(inbox)

# SHOW_LABEL(lid): done
# - get mid... from label database (lid -> mid...)
# - LIST_INDEX(mid...)

# LIST_INDEX(mid..., thread): done - list.rb
# - get index line from index database
# - if thread:
#   + sort by thread, date
#   + list with thread indenting
# - else: list

# SHOW_THREAD(mid): done
# - get tid from mid
# - get mid... from tid
# - LIST_INDEX(mid...)

# SHOW_QUERY(query): done
# - get qresult... from query using inverted index
# - LIST_INDEX(qresult..., false)

# SHOW_SIMILARITY(mid):
# - get most significant words from mid using inverted index
# - get qresult... from words using inverted index
# - trim qresult... using score
# - LIST_INDEX(qresult..., thread=false)

# READ(mid): done
# - MARK_READ(mid)
# - return path (mid[3]/mid[0]/mid)

# MARK_READ_ON/MARK_FLAG_ON(mid): done - action.rb
# - get index(mid)
# - index->flags |= FLAG (READ or FLAG)
# - upadte index

# MARK_READ_OFF/MARK_FLAG_OFF(mid): done - action.rb
# - get index(mid)
# - index->flags &= ~FLAG (READ or FLAG)
# - upadte index

# ARCHIVE_MAIL(mid): done - action.rb
# - REMOVE_LABEL(mid, inbox)
# - remove mid from inbox

# ARCHIVE_THREAD(tid): done - action.rb
# - get mid.. from tid
# - REMOVE_MAIL(mid, inbox) for each mid...

# DELETE(mid): done - action.rb
# - remove mid from index database
# - remove mid from label database
# - remove mid from thread database
# - remove file

# ADD_LABEL(mid, label): done - action.rb
# - get lid (label -> lid); generate if none
# - add mid to (lid -> mid...)
# - add lid to (mid -> lid...)

# REMOVE_LABEL(mid, label): done - action.rb
# - get lid (label -> lid); return if none
# - delete mid from (lid -> mid...)
# - delete lid from (mid -> lid...)

# KILL_LABEL(label): done - action.rb
# - get lid (label -> lid); return if none
# - delete mid from (lid -> mid...)
# - delete lid from (mid -> lid...)

# INBOX(mid): done
# - ADD_LABEL(mid, inbox)

# TRASH(mid): done
# - REMOVE_LABEL(mid, inbox)
# - ADD_LABEL(mid, trash)

# SPAM(mid): done
# - REMOVE_LABEL(mid, inbox)
# - ADD_LABEL(mid, spam)
# - update spam database (+add)

# NO_SPAM(mid): done
# - REMOVE_LABEL(mid, spam)
# - update spam database (-add)
# - ADD_LABEL(mid, inbox)

# CLEAR_TRASH: done
# - get mid... from (lid(=trash) -> mid...)
# - for mid in mid...; DELETE(mid)
# - get mid... from (lid(=spam) -> mid...)
# - for mid in mid...; DELETE(mid)

# COMPACT
# - estindex purge
# - estindex optimize
# - compact index database (mid -> flags,..)
#   + delete if each mid exists
# - compact label database (lid -> mid...)
#   + delete if each mid exists
# - compact label database (mid -> lid...)
#   + delete if each mid exists
# - compact thread database (message-id -> mid)
#   + delete if each mid exists
# - compact thread database (mid -> tid)
#   + delete if each mid exists
# - compact thread database (tid -> mid)
#   + delete if each mid exists
# - compact thread database (tid -> mid)
#   + delete if each mid exists
# - update thread database

# ---- DB

# index database
# - mid -> flags, tid, date, from, subject

# label database
# - label -> lid
# - lid -> mid...
# - mid -> lid...

# thread database
# - message-id -> mid
# - mid -> tid
# - tid -> mid...

# spam database
# - externel spam filter

